//-----------------------------------------------------------------------
//  This file is part of the Microsoft Robotics Studio Code Samples.
//
//  Copyright (C) Microsoft Corporation.  All rights reserved.
//
//  $File: SimulatedLRF.cs $ $Revision: 22 $
//-----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using permissions = System.Security.Permissions;
using svcbase = Microsoft.Dss.ServiceModel.DsspServiceBase;
using dssp = Microsoft.Dss.ServiceModel.Dssp;
using sicklrf = Microsoft.Robotics.Services.Sensors.SickLRF.Proxy;
using submgr = Microsoft.Dss.Services.SubscriptionManager;
using simtypes = Microsoft.Robotics.Simulation;
using simengine = Microsoft.Robotics.Simulation.Engine;
using physics = Microsoft.Robotics.Simulation.Physics;
using Microsoft.Robotics.PhysicalModel;
using System.ComponentModel;
using Microsoft.Dss.Core.DsspHttp;
using System.Net;
using Microsoft.Dss.Core.DsspHttpUtilities;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using W3C.Soap;
using Microsoft.Dss.ServiceModel.Dssp;
using System.Net.Mime;
using System.Drawing.Drawing2D;

namespace Microsoft.Robotics.Services.Simulation.Sensors.LaserRangeFinder
{
    /// <summary>
    /// Provides access to a simulated Laser Range Finder contract 
    /// using physics raycasting and the LaserRangeFinderEntity
    /// </summary>
    [DisplayName("Simulated Laser Range Finder")]
    [Description("Provides access to a simulated laser range finder.\n(Uses the Sick Laser Range Finder contract.)")]
    [AlternateContract(sicklrf.Contract.Identifier)]
    [Contract(Contract.Identifier)]
    public class SimulatedLRFService : svcbase.DsspServiceBase
    {
        #region Simulation Variables
        simengine.LaserRangeFinderEntity _entity;
        simengine.SimulationEnginePort _notificationTarget;
        physics.PhysicsEngine _physicsEngine;
        Port<physics.RaycastResult> _raycastResults = new Port<physics.RaycastResult>();
        #endregion

        private sicklrf.State _state = new sicklrf.State();
     
        [ServicePort("/SimulatedLRF", AllowMultipleInstances=true)]
        private sicklrf.SickLRFOperations _mainPort = new sicklrf.SickLRFOperations();

        [Partner("SubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways)]
        submgr.SubscriptionManagerPort _subMgrPort = new submgr.SubscriptionManagerPort();

        //xslt
        [EmbeddedResource("Microsoft.Robotics.Services.LaserRangeFinder.Simulated.DssSimulatedLRF.SickLRF.xslt")]
        string _transform = null;
        DsspHttpUtilitiesPort _httpUtilities;

        private static readonly int robotWidth = 400; //mm
        private static readonly int maxDist = 8192; //mm


        public SimulatedLRFService(dssp.DsspServiceCreationPort creationPort) : 
                base(creationPort)
        {
        }

        protected override void Start()
        {
            _physicsEngine = physics.PhysicsEngine.GlobalInstance;
            _notificationTarget = new simengine.SimulationEnginePort();

            // PartnerType.Service is the entity instance name. 
            simengine.SimulationEngine.GlobalInstancePort.Subscribe(ServiceInfo.PartnerList, _notificationTarget);

            //xslt
            _httpUtilities = DsspHttpUtilitiesService.Create(Environment);

            // dont start listening to DSSP operations, other than drop, until notification of entity
            Activate(new Interleave(
                new TeardownReceiverGroup
                (
                    Arbiter.Receive<simengine.InsertSimulationEntity>(false, _notificationTarget, InsertEntityNotificationHandlerFirstTime),
                    Arbiter.Receive<dssp.DsspDefaultDrop>(false, _mainPort, DefaultDropHandler)
                ),
                new ExclusiveReceiverGroup(),
                new ConcurrentReceiverGroup()
            ));

        }

        const float LASER_RANGE = 8f;
        void CreateDefaultState()
        {
            _state.Units = sicklrf.Units.Millimeters;
            _state.AngularRange = 180;
            _state.AngularResolution = 0.5f;
        }

        void DeleteEntityNotificationHandler(simengine.DeleteSimulationEntity del)
        {
            _entity = null;
        }

        void InsertEntityNotificationHandlerFirstTime(simengine.InsertSimulationEntity ins)
        {
            InsertEntityNotificationHandler(ins);
            base.Start();
            MainPortInterleave.CombineWith(
                new Interleave(
                    new TeardownReceiverGroup(),
                    new ExclusiveReceiverGroup(
                        Arbiter.Receive<simengine.InsertSimulationEntity>(true, _notificationTarget, InsertEntityNotificationHandler),
                        Arbiter.Receive<simengine.DeleteSimulationEntity>(true, _notificationTarget, DeleteEntityNotificationHandler)
                    ),
                    new ConcurrentReceiverGroup()
                )
            );
        }

        void InsertEntityNotificationHandler(simengine.InsertSimulationEntity ins)
        {
            _entity = (simengine.LaserRangeFinderEntity)ins.Body;
            _entity.ServiceContract = Contract.Identifier;

            CreateDefaultState();

            physics.RaycastProperties raycastProperties = new physics.RaycastProperties();
            raycastProperties.StartAngle = -_state.AngularRange / 2.0f;
            raycastProperties.EndAngle = _state.AngularRange / 2.0f;
            raycastProperties.AngleIncrement = (float)_state.AngularResolution;
            raycastProperties.Range = LASER_RANGE;
            raycastProperties.OriginPose = new Pose();

            _entity.RaycastProperties = raycastProperties;
            _entity.Register(_raycastResults);

            // attach handler to raycast results port
            Activate(Arbiter.Receive(true, _raycastResults, RaycastResultsHandler));
        }

        private void RaycastResultsHandler(physics.RaycastResult result)
        {
            // we just receive ray cast information from physics. Currently we just use
            // the distance measurement for each impact point reported. However, our simulation
            // engine also provides you the material properties so you can decide here to simulate
            // scattering, reflections, noise etc.
            
            sicklrf.State latestResults = new sicklrf.State();
            latestResults.DistanceMeasurements = new int[result.SampleCount+1];
            latestResults.DistanceMeasurements.SetValue((int)(LASER_RANGE*1000f),0);
            foreach (physics.RaycastImpactPoint pt in result.ImpactPoints)
            {
                // the distance to the impact has been pre-calculted from the origin 
                // and it's in the fourth element of the vector
                latestResults.DistanceMeasurements[pt.ReadingIndex] = (int)(pt.Position.W * 1000f);
            }

            latestResults.AngularRange = (int)Math.Abs(_entity.RaycastProperties.EndAngle - _entity.RaycastProperties.StartAngle);
            latestResults.AngularResolution = (int)_entity.RaycastProperties.AngleIncrement;
            latestResults.Units = sicklrf.Units.Millimeters;
            latestResults.LinkState = "Measurement received";
            latestResults.TimeStamp = DateTime.Now;
            
            // send replace message to self
            sicklrf.Replace replace = new sicklrf.Replace();
            // for perf reasons dont set response port, we are just talking to ourself anyway
            replace.ResponsePort = null;
            replace.Body = latestResults;
            _mainPort.Post(replace);
        }

        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> GetHandler(sicklrf.Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReplaceHandler(sicklrf.Replace replace)
        {
            _state = replace.Body;
            if (replace.ResponsePort != null)
                replace.ResponsePort.Post(dssp.DefaultReplaceResponseType.Instance);

            // issue notification
            _subMgrPort.Post(new submgr.Submit(_state, dssp.DsspActions.ReplaceRequest));
            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> SubscribeHandler(sicklrf.Subscribe subscribe)
        {
            yield return Arbiter.Choice(
                SubscribeHelper(_subMgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    _subMgrPort.Post(new submgr.Submit(
                        subscribe.Body.Subscriber, dssp.DsspActions.ReplaceRequest, _state, null));
                },
                delegate(Exception ex) { LogError(ex); }
            );
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReliableSubscribeHandler(sicklrf.ReliableSubscribe subscribe)
        {
            yield return Arbiter.Choice(
                SubscribeHelper(_subMgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    _subMgrPort.Post(new submgr.Submit(
                        subscribe.Body.Subscriber, dssp.DsspActions.ReplaceRequest, _state, null));
                },
                delegate(Exception ex) { LogError(ex); }
            );
        }

        #region HttpGet Handlers

        //[ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        //public IEnumerator<ITask> GetHandler(HttpGet get)
        //{
        //    get.ResponsePort.Post(new HttpResponseType(HttpStatusCode.OK, _state));
        //    yield break;
        //}

        static readonly string _root = "/sicklrf";
        static readonly string _cylinder = "/cylinder";
        static readonly string _top = "/top";
        static readonly string _raw = "/raw";

        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public void HttpGetHandler(HttpGet httpGet)
        {
            HttpListenerRequest request = httpGet.Body.Context.Request;
            HttpListenerResponse response = httpGet.Body.Context.Response;

            Stream image = null;

            string path = request.Url.AbsolutePath;

            if (path.EndsWith(_cylinder))
            {
                image = GenerateCylinder(robotWidth);
            }
            else if (path.EndsWith(_top))
            {
                image = GenerateTop(robotWidth);
            }
            else if (path.EndsWith(_raw))
            {
                //no transform
                HttpResponseType rsp = new HttpResponseType(HttpStatusCode.OK, _state);
                httpGet.ResponsePort.Post(rsp);
            }
            else if (path.EndsWith(_root))
            {
                HttpResponseType rsp = new HttpResponseType(HttpStatusCode.OK,
                _state,
                _transform);
                httpGet.ResponsePort.Post(rsp);
            }



            if (image != null)
            {
                SendJpeg(httpGet.Body.Context, image);
            }
            else
            {
                httpGet.ResponsePort.Post(Fault.FromCodeSubcodeReason(
                    W3C.Soap.FaultCodes.Receiver,
                    DsspFaultCodes.OperationFailed,
                    "Unable to generate Image"));
            }
        }

        private void SendJpeg(HttpListenerContext context, Stream stream)
        {
            WriteResponseFromStream write = new WriteResponseFromStream(context, stream, MediaTypeNames.Image.Jpeg);

            _httpUtilities.Post(write);

            Activate(
                Arbiter.Choice(
                    write.ResultPort,
                    delegate(Stream res)
                    {
                        stream.Close();
                    },
                    delegate(Exception e)
                    {
                        stream.Close();
                        LogError(e);
                    }
                )
            );
        }

        #endregion

        #region Image generators

        private Stream GenerateCylinder(int width)
        {
            if (_state.DistanceMeasurements == null)
            {
                return null;
            }

            MemoryStream memory = null;

            using (Bitmap bmp = new Bitmap(_state.DistanceMeasurements.Length, 100))
            {
                using (Graphics g = Graphics.FromImage(bmp))
                {
                    g.Clear(Color.LightGray);

                    int half = bmp.Height / 2;
                    int middle = _state.DistanceMeasurements.Length / 2;

                    for (int x = 0; x < _state.DistanceMeasurements.Length; x++)
                    {
                        int range = _state.DistanceMeasurements[x];
                        if (range > 0 && range < maxDist)
                        {
                            int h = bmp.Height * 500 / _state.DistanceMeasurements[x];
                            if (h < 0)
                            {
                                h = 0;
                            }

                            double angle = x * Math.PI * _state.AngularRange / _state.DistanceMeasurements.Length / 180; //radians
                            double obsThresh = Math.Abs(width / (2 * Math.Cos(angle)));

                            Color col;
                            if (range < obsThresh)
                                col = LinearColor(Color.DarkRed, Color.LightGray, 0, maxDist, range);
                            else
                                col = LinearColor(Color.DarkBlue, Color.LightGray, 0, maxDist, range);

                            g.DrawLine(new Pen(col), bmp.Width - x, half - h, bmp.Width - x, half + h);
                        }
                    } 
                    g.DrawString(
                        _state.TimeStamp.ToString(), 
                        new Font(FontFamily.GenericSansSerif, 10, GraphicsUnit.Pixel), 
                        Brushes.Gray, 0, 0
                    );
                }

                memory = new MemoryStream();
                bmp.Save(memory, ImageFormat.Jpeg);
                memory.Position = 0;
            }

            return memory;
        }

        private Color LinearColor(Color nearColor, Color farColor, int nearLimit, int farLimit, int value)
        {
            if (value <= nearLimit)
            {
                return nearColor;
            }
            else if (value >= farLimit)
            {
                return farColor;
            }

            int span = farLimit - nearLimit;
            int pos = value - nearLimit;

            int r = (nearColor.R * (span - pos) + farColor.R * pos) / span;
            int g = (nearColor.G * (span - pos) + farColor.G * pos) / span;
            int b = (nearColor.B * (span - pos) + farColor.B * pos) / span;

            return Color.FromArgb(r, g, b);
        }

        private Stream GenerateTop(int width)
        {
            if (_state.DistanceMeasurements == null)
            {
                return null;
            }

            MemoryStream memory = null;

            int height = width / 2;
            using (Bitmap bmp = new Bitmap(width, height))
            {
                using (Graphics g = Graphics.FromImage(bmp))
                {
                    g.Clear(Color.LightGray);

                    double angularOffset = -90 + _state.AngularRange / 2.0;
                    double piBy180 = Math.PI / 180.0;
                    double halfAngle = _state.AngularRange / 360.0 / 2.0;
                    double scale = (double)height / (double)maxDist;

                    GraphicsPath path = new GraphicsPath();

                    for (int pass = 0; pass != 2; pass++)
                    {
                        for (int i = 0; i < _state.DistanceMeasurements.Length; i++)
                        {
                            int range = _state.DistanceMeasurements[i];

                            if (range <= 0)
                                range = maxDist;

                            //if (range > 0 && range < maxDist)
                            //{
                                double angle = i * (_state.AngularRange / 360.0) - angularOffset;
                                double lowAngle = (angle - halfAngle) * piBy180;
                                double highAngle = (angle + halfAngle) * piBy180;

                                double drange = range * scale;

                                float lx = (float)(height + drange * Math.Cos(lowAngle));
                                float ly = (float)(height - drange * Math.Sin(lowAngle));
                                float hx = (float)(height + drange * Math.Cos(highAngle));
                                float hy = (float)(height - drange * Math.Sin(highAngle));

                                if (pass == 0)
                                {
                                    if (i == 0)
                                    {
                                        path.AddLine(height, height, lx, ly);
                                    }
                                    path.AddLine(lx, ly, hx, hy);
                                }
                                else
                                {
                                    if (range < maxDist)
                                        g.DrawLine(Pens.DarkBlue, lx, ly, hx, hy);
                                }
                            //}
                        }

                        if (pass == 0)
                        {
                            g.FillPath(Brushes.White, path);
                        }
                    }

                    float botWidth = (float)((robotWidth/2.0) * scale);
                    g.DrawLine(Pens.Red, height, height - botWidth, height, height);
                    g.DrawLine(Pens.Red, height - 3, height - botWidth, height + 3, height - botWidth);
                    g.DrawLine(Pens.Red, height - botWidth, height - 3, height - botWidth, height);
                    g.DrawLine(Pens.Red, height + botWidth, height - 3, height + botWidth, height);
                    g.DrawLine(Pens.Red, height - botWidth, height - 1, height + botWidth, height - 1);

                    g.DrawString(
                        _state.TimeStamp.ToString(),
                        new Font(FontFamily.GenericSansSerif, 10, GraphicsUnit.Pixel),
                        Brushes.Gray, 0, 0
                    );
                }

                memory = new MemoryStream();
                bmp.Save(memory, ImageFormat.Jpeg);
                memory.Position = 0;

            }
            return memory;
        }

        #endregion

    }
}
